//
//	File:			Brain.java
//	Author:		Krzysztof Langner
//	Date:			1997/04/28
//

import java.lang.Math;
import java.util.StringTokenizer;

import java.awt.*;

class Brain extends Thread implements SensorInput
{
   private final int ACTION_DO_NOTHING=0;
   private final int ACTION_LOOK_AROUND=1;
   private final int ACTION_GO_TO_POS=2;
   private final int ACTION_GO_HOME=3;
   private final int ACTION_LOOK_TO_BALL=4;
   private final int ACTION_KICK_BALL=5;
   private final int ACTION_GO_TO_BALL=6;

   private final int ACTION_POBEHUJ_DO_STVORCA=7;

   private final String action_names[]= {"Do Nothing",
                                                "Look around",
                                                "Go To Position",
                                                "Go Home",
                                                "Look to Ball",
                                                "Kick Ball",
                                                "Go to Ball",
                                                "Pobehuj do stvorca"};

   private final int ACTION_DEFAULT_ACTION = ACTION_GO_HOME;

   private int current_action=0;

   //position where krislet nedd to go...
   //it is used in action_go_to_pos method
   private Pos endPos = new Pos();
   private Pos currentPos = new Pos();

   private ObjectInfo object;
   private ObjectInfo ball = new BallInfo();

   //static public ObjectInfo staticBall = new StaticBallInfo();

   private Pos currentBallPos = new Pos();
   private Pos lastBallPos = new Pos();

   private Rectangle home_area;
   private Pos home_pos;
   private Pos start_pos;

   private double tmpUhol=0;

   //typ hraca
   int type=0;


	//---------------------------------------------------------------------------
	// This constructor:
	// - stores connection to krislet
	// - starts thread for this object
	public Brain(SendCommand krislet, String team,
							char side, int number, String playMode)
	{
		m_timeOver = false;
		m_krislet = krislet;
		m_memory = new Memory(side);
//		m_team = team;
		m_side = side;
//		m_number = number;
//		m_playMode = playMode;


      type=krislet.getType();
      if (type == Krislet.TYPE_LEFT_BACK)
      {
         home_area=new Rectangle(-30, -40, 30, 40);
         home_pos = new Pos(-8,-3);
         start_pos = new Pos(-10,-5);
      }
      else if (type == Krislet.TYPE_LEFT_FRONT)
      {
         home_area=new Rectangle(0, -40, 30, 40);
         home_pos = new Pos(10, -10);
         start_pos = new Pos(-3,-10);
      }
      else if (type == Krislet.TYPE_RIGHT_BACK)
      {
         home_area=new Rectangle(-30, 0, 30, 50);
         home_pos = new Pos(-15,15);
         start_pos = new Pos(-10,5);
      }
      else if (type == Krislet.TYPE_RIGHT_FRONT)
      {
         home_area=new Rectangle(0, 0, 30, 40);
         home_pos = new Pos(10, 10);
         start_pos = new Pos(-3, 10);
      }
      else if (type == Krislet.TYPE_CENTER_BACK)
      {
         home_area=new Rectangle(-50, -20, 20, 40);
         home_pos = new Pos(-40, 0);
         start_pos = new Pos(-40, 0);
      }
      else if (type == Krislet.TYPE_LEFT_MIDDLE)
      {
         home_area=new Rectangle(-30, -40, 60, 40);
         home_pos = new Pos(-10,-3);
         start_pos = new Pos(-10,-3);
      }
      else if (type == Krislet.TYPE_RIGHT_MIDDLE)
      {
         home_area=new Rectangle(-30, 0, 60, 40);
         home_pos = new Pos(-10,3);
         start_pos = new Pos(-10,3);
      }
      else
      {
         debug("Unknown player type: "+type);
         home_pos = new Pos(0,0);
         start_pos = new Pos(0,0);
      }
      current_action = ACTION_LOOK_TO_BALL;

      debug("Playing as: "+type+" home area="+home_area.getX()+","+home_area.getY()+","+home_area.getWidth()+","+home_area.getHeight());

		start();
	}


	//---------------------------------------------------------------------------
	// This is main brain function used to make decision
	// In each cycle we decide which command to issue based on
	// current situation. the rules are:
	//
	//	1. If you don't know where is ball then turn right and wait for new info
	//
	//	2. If ball is too far to kick it then
	//		2.1. If we are directed towards the ball then go to the ball
	//		2.2. else turn to the ball
	//
	//	3. If we dont know where is opponent goal then turn wait
	//				and wait for new info
	//
	//	4. Kick ball
	//
	//	To ensure that we don't send commands to often after each cycle
	//	we waits one simulator steps. (This of course should be done better)
	public void run()
	{
		// first put it somewhere on my side
		//m_krislet.move( -Math.random()*52.5 , Math.random()*34.0 );
      waitForNextSend();
      m_krislet.move(start_pos.x, start_pos.y);
      //m_memory.setCurrentPos(new Pos(-20,5));
      current_action = ACTION_LOOK_TO_BALL;

      //object = m_memory.getObject("ball");
      //m_krislet.turn();

      boolean isObjectInHomeArea=false;
      boolean ballWasSennLastTime=true;

		while( !m_timeOver )
		{
         m_memory.waitForNewInfo();
         lastKrisletBodyDir = Krislet.bodyDir;
         m_memory.estimateCurrentPosition();
         currentPos = m_memory.getCurrentPos();
         debug("--Position: x="+(int)currentPos.x+" y="+(int)currentPos.y+" m_side="+m_side+" play_mode="+play_mode+" time="+Krislet.server_time);

         ObjectInfo o_ball = m_memory.getObject("ball");
         if (o_ball!=null)
         {
            ball = new BallInfo();
            ball.m_direction = o_ball.m_direction;
            ball.m_distance = o_ball.m_distance;

            if (Math.abs(o_ball.m_direction)>45.0)
            {
               //zachovaj stare info o zmene, nove nie je k dispozicii
            }
            else
            {
               ball.m_dirChange = o_ball.m_dirChange;
               ball.m_distChange = o_ball.m_distChange;
            }

            lastBallDir = o_ball.m_direction;

            lastBallPos.x = currentBallPos.x;
            lastBallPos.y = currentBallPos.y;

            currentBallPos = m_memory.estimateObjectPosition(ball);

            estimateNextStepBallPosition();
            ballWasSennLastTime=true;
         }
         else
         {
            //TOTO TREBA PREROBIT!!!!

            //ak si naposledy videl loptu, predpokladaj ze ju este vidis
            if (ballWasSennLastTime==false) ball=null;

            ballWasSennLastTime=false;

            lastBallPos.x = -999;
            currentBallPos.x = -999;

         }

         //posunutie hracov - LEFT MIDDLE
         if (type == Krislet.TYPE_LEFT_MIDDLE)
         {
            if (currentBallPos.x < -5)
            {
               if (currentBallPos.x > -50)
               {
                  home_pos.x = ((currentBallPos.x+52.5) / 2.00) -52.5;
               }
               if (ball!=null)
               {
                  home_pos.y = (currentBallPos.y / (currentBallPos.x + 52.5)) * (home_pos.x+52.5);
               }
               debug("Menim home pos1: "+home_pos.x+","+home_pos.y+" ball="+currentBallPos.x+","+currentBallPos.y);
            }
            else
            {
               home_pos.x = start_pos.x;
               home_pos.y = start_pos.y;
               debug("Menim home pos2: "+home_pos.x+","+home_pos.y+" ball="+currentBallPos.x+","+currentBallPos.y);
            }

         }


         //posunutie hracov - RIGHT MIDDLE
         if (type == Krislet.TYPE_RIGHT_MIDDLE)
         {
           if (currentBallPos.x < -5)
            {
               if (currentBallPos.x > -50)
               {
                  home_pos.x = ((currentBallPos.x+52.5) / 2.00) -52.5;
               }
               if (ball!=null)
               {
                  home_pos.y = (currentBallPos.y / (currentBallPos.x + 52.5)) * (home_pos.x+52.5);
               }
               debug("Menim home pos1: "+home_pos.x+","+home_pos.y+" ball="+currentBallPos.x+","+currentBallPos.y);
            }
            else
            {
               home_pos.x = start_pos.x;
               home_pos.y = start_pos.y;
               debug("Menim home pos2: "+home_pos.x+","+home_pos.y+" ball="+currentBallPos.x+","+currentBallPos.y);
            }
         }

         //posunutie hracov - CENTER BACK
         if (type == Krislet.TYPE_CENTER_BACK && currentBallPos.x > -15)
         {
            home_pos.x = -40;
            if (ball!=null)
            {
               home_pos.y = (currentBallPos.y / (currentBallPos.x + 52.5)) * (home_pos.x+52.5);
            }
         }

         if (ball!=null)
         {

            //debug("--- SEE BALL !!! ---");
            isObjectInHomeArea = isObjectInHomeArea(ball);
            if (isObjectInHomeArea) debug("Ball is in myHomeArea");
            if ( ball.m_distance < 1.0)
            {
               current_action = ACTION_KICK_BALL;
            }
            else if (isObjectInHomeArea)
            {
               current_action = ACTION_GO_TO_BALL;
            }
            else if (current_action != ACTION_GO_TO_BALL && currentPos.distance(home_pos) > 2.0)
            {
               current_action = ACTION_GO_HOME;
            }
            /*else
            {
               current_action = ACTION_LOOK_TO_BALL;
            } */

            debug("BASIC LOGIC - action="+current_action);

            if (current_action != ACTION_KICK_BALL)
            {
               //ak je niekto blizsie, vypaz sa na loptu
               if (m_memory.isSomeoneNear(ball))
               {
                  //if (current_action!=ACTION_LOOK_TO_BALL) current_action = ACTION_GO_HOME;
                  //else current_action = ACTION_LOOK_TO_BALL;
                  if (isObjectInHomeArea)
                  {
                     current_action = ACTION_GO_TO_BALL;
                  }
                  else
                  {
                     current_action = ACTION_GO_HOME;
                  }
               }
               else
               {
                  //if (current_action == ACTION_GO_HOME) current_action = ACTION_GO_TO_BALL;
                  current_action = ACTION_GO_TO_BALL;
               }
            }

            debug("BASIC LOGIC2 - action="+current_action);

            if (current_action==ACTION_GO_HOME && currentPos.distance(home_pos) <=2.0)
            {
               debug("Som dostatocne doma...");
               current_action = ACTION_LOOK_TO_BALL;
            }

            debug("BASIC LOGIC3 - action="+current_action);
         }
         else
         {
            //debug("--- CAN'T SEE BALL !!! ---");
            if (currentPos.distance(home_pos) > 2.0 && current_action!=ACTION_LOOK_TO_BALL && current_action!=ACTION_GO_TO_BALL)  current_action = ACTION_GO_HOME;
            else current_action = ACTION_LOOK_TO_BALL;
         }

         if (!m_memory.isPositionEstimated()) current_action = ACTION_LOOK_AROUND;

         if (current_action == ACTION_GO_HOME)
         {
            if (goHomeCounter >= goHomeMaxSteps)
            {
               goHomeCounter=0;
               current_action = ACTION_LOOK_TO_BALL;

               if (Krislet.bodyDir > 45) lastBallDir = -10;
               if (Krislet.bodyDir < -45) lastBallDir = 10;

            }
         }

         if (play_mode!=PLAY_ON && current_action != ACTION_KICK_BALL) current_action = ACTION_LOOK_TO_BALL;
         if (play_mode!=PLAY_ON && type==Krislet.TYPE_LEFT_MIDDLE && current_action != ACTION_KICK_BALL)
         {
            //ak som lavy hrac
            if (m_side=='l')
            {
               current_action = ACTION_LOOK_TO_BALL;
               if (play_mode==KICK_OFF_L || play_mode==KICK_IN_L || play_mode==BEFORE_KICK_OFF ||
                   play_mode==CORNER_KICK_L || play_mode==GOAL_KICK_L || play_mode==KICK_BALL_LINE)
               {
                  current_action=ACTION_GO_TO_BALL;
               }

            }
            else
            {
               current_action = ACTION_LOOK_TO_BALL;
               if (play_mode==KICK_OFF_R || play_mode==KICK_IN_R ||
                   play_mode==CORNER_KICK_R || play_mode==GOAL_KICK_R || play_mode==KICK_BALL_LINE)
               {
                  current_action=ACTION_GO_TO_BALL;
               }
            }



         }

         debug("Current action="+action_names[current_action]);
         switch (current_action)
         {
            case ACTION_DO_NOTHING:
               action_do_nothing();
               break;
            case ACTION_LOOK_AROUND:
               action_look_around();
               break;
            case ACTION_GO_TO_POS:
               action_go_to_pos();
               break;
            case ACTION_GO_HOME:
               action_go_home();
               break;
            case ACTION_LOOK_TO_BALL:
               action_look_to_ball();
               break;
            case ACTION_GO_TO_BALL:
               action_go_to_ball();
               break;
            case ACTION_KICK_BALL:
               action_kick_ball();
               break;
            case ACTION_POBEHUJ_DO_STVORCA:
               action_pobehuj_do_stvorca();
               break;
            default:
               action_do_nothing();
         }

		}

	}


//===========================================================================
// Here are suporting functions for implement logic

   private void action_do_nothing()
   {
      current_action = ACTION_LOOK_TO_BALL;

      return;
   }

   private void old_logic()
   {

  			object = m_memory.getObject("ball");
         object=null;
			if( object == null )
			{
            //if (waitForNextSend()==true) return;
            goToPos(new Pos(32,-10));
			}
			else if( object.m_distance > 1.5 )
			{
				// If ball is too far then
				// turn to ball or
				// if we have correct direction then go to ball
            if (waitForNextSend()==true) return;
				if( object.m_direction != 0 )
					m_krislet.turn(object.m_direction);
				else
					m_krislet.dashToDist(object.m_distance);
			}
			else
			{
				// We know where is ball and we can kick it
				// so look for goal
				if( m_side == 'l' )
					object = m_memory.getObject("goal r");
				else
					object = m_memory.getObject("goal l");

				if( object == null )
				{
               if (waitForNextSend()==true) return;
					m_krislet.turn(60);
					m_memory.waitForNewInfo();
				}
				else
            {
               if (waitForNextSend()==true) return;
					m_krislet.kick(100, object.m_direction);
            }
			}
   }

   private void action_look_around()
   {
      current_action = ACTION_LOOK_AROUND;
      //TODO: dorobit lepsie obzeranie (aby sa tocil spravnym smerom)
      //if (waitForNextSend()==true) return;
      if (lastBallDir > 0) m_krislet.turn(60);
      else m_krislet.turn(-60);
   }

   private void action_go_to_pos()
   {
      current_action = ACTION_GO_TO_POS;
      if (currentPos.distance(endPos) > 2.0)
      {
         //if (waitForNextSend()==true) return;
         goToPos(endPos);
      }
      else current_action = ACTION_LOOK_TO_BALL;
   }

   private void action_go_home()
   {
      current_action = ACTION_GO_HOME;



      endPos.x = home_pos.x;
      endPos.y = home_pos.y;
      if (Krislet.pos.distance(home_pos) > 2.0)
      {
         //if (waitForNextSend()==true) return;
         debug("Going home: "+home_pos.x+","+home_pos.y);
         goHomeCounter++;
         goToPos(home_pos);
      }
      else
      {
         debug("Som dost blizko home: "+home_pos.x+","+home_pos.y);
         current_action = ACTION_LOOK_TO_BALL;
      }
   }

   private void action_look_to_ball()
   {
      current_action = ACTION_LOOK_TO_BALL;

		if( ball == null )
		{
         debug("BALL: can't see it...");
         //if (waitForNextSend()==true) return;

         if (lastBallDir > 0) m_krislet.turn(50);
         else m_krislet.turn(-50);
		}
      //ignoruj presne natocenie
		else if( Math.abs(ball.m_direction) > 20.00)
      {
         debug("BALL: yeah, i see ball: angle="+ball.m_direction+" dist="+ball.m_distance);
         //if (waitForNextSend()==true) return;
			m_krislet.turn(ball.m_direction);
		}
   }

   private void action_kick_ball()
   {
      current_action = ACTION_KICK_BALL;
      if( m_side == 'l' )
			object = m_memory.getObject("goal r");
		else
			object = m_memory.getObject("goal l");

      PlayerInfo forward = m_memory.findNearestForward();

      //ak sa este nehra
      if (play_mode!=PLAY_ON)
      {
         if (play_mode==BEFORE_KICK_OFF || play_mode==KICK_OFF_L || play_mode==KICK_OFF_R)
         {
            m_krislet.kick(100, 40 - Krislet.bodyDir);
            return;
         }

      }

		if( forward == null )
		{
         //ak som daleko od brany sa otoc a strel
         if (Krislet.pos.x < 15)
         {
            double dir = -Krislet.bodyDir;
            if (Math.abs(dir) > 90)
            {
               if (dir < 0) dir = -90;
               else dir = 90;
            }
            //if (waitForNextSend()==true) return;
            dir = m_memory.findClearAngleForBall(dir);
            m_krislet.kick(80, dir);
            if (waitForNextSend()==true) return;
            if (Math.abs(ball.m_direction - dir) < 15)
            {
			      m_krislet.turn(dir);
               if (waitForNextSend()==true) return;
               m_krislet.dashToDist(80);
            }
            else
            {
               m_krislet.turn(ball.m_direction);
            }
         }
         else
         {
            if (object==null)
            {
               //if (waitForNextSend()==true) return;
			      //m_krislet.turn(60);

               double goal_pos_y=0;
               if (Krislet.pos.y > 0) goal_pos_y = 3;
               else goal_pos_y = -3;

               double dx= 52.5 - Krislet.pos.x;
               double dy = goal_pos_y - Krislet.pos.y;
               double adir = Math.toDegrees(Math.atan2(dy, dx));

               double dir = -Krislet.bodyDir + adir;

               if (Math.abs(dir) > 90)
               {
                  if (dir < 0) dir = -90;
                  else dir = 90;
               }
               //if (waitForNextSend()==true) return;
               dir = m_memory.findClearAngleForBall(dir);
               m_krislet.kick(80, dir);
               if (waitForNextSend()==true) return;
			      if (Math.abs(ball.m_direction - dir) < 15)
               {
			         m_krislet.turn(dir);
                  if (waitForNextSend()==true) return;
                  m_krislet.dashToDist(80);
               }
               else
               {
                  m_krislet.turn(ball.m_direction);
               }

            }
            else
            {
               //if (waitForNextSend()==true) return;
               //kick to goal
               double dir;
               dir = object.m_direction;
               dir = m_memory.findClearAngleForBall(dir);
               m_krislet.kickToDist(100, dir);
               if (waitForNextSend()==true) return;

               if (Math.abs(ball.m_direction - dir) < 15)
               {
			         m_krislet.turn(object.m_direction);
                  if (waitForNextSend()==true) return;
                  m_krislet.dashToDist(100);
               }
               else
               {
                  m_krislet.turn(ball.m_direction);
               }
            }
         }
         current_action = ACTION_GO_TO_BALL;
         return;
		}
      //mam svojho hraca
		else
      {
         double clear_angle = m_memory.findClearAngle(forward.m_direction);

         clear_angle = m_memory.findClearAngleForBall(clear_angle);

         if (Math.abs(clear_angle - forward.m_direction) > 5)
         {
            debug("---->NASIEL SOM UTOCNIKA, ale je tam niekto cudzi");
            m_krislet.kick(40, clear_angle - forward.m_direction);
            if (waitForNextSend()==true) return;

            if (Math.abs(ball.m_direction - clear_angle - forward.m_direction) < 15)
            {
               m_krislet.turn(clear_angle - forward.m_direction);
               if (waitForNextSend()==true) return;
               m_krislet.dash(40);
            }
            else
            {
               m_krislet.turn(ball.m_direction);
            }


         }
         else
         {


               debug("---->NASIEL SOM UTOCNIKA!!!!");
               //if (waitForNextSend()==true) return;
               m_krislet.kickToDist(forward.m_distance, forward.m_direction);
               if (waitForNextSend()==true) return;

               if (Math.abs(ball.m_direction - forward.m_direction) < 15)
               {
                  m_krislet.turn(forward.m_direction);
                  if (waitForNextSend()==true) return;
                  m_krislet.dashToDist(forward.m_distance);
               }
               else
               {
                  m_krislet.turn(ball.m_direction);
               }
         }
      }

      //if (Krislet.pos.x > FORWARD_LINE) current_action = ACTION_GO_HOME;
      current_action = ACTION_GO_TO_BALL;
   }

   private void action_go_to_ball()
   {
      current_action = ACTION_GO_TO_BALL;
      if (ball == null)
      {
         action_look_to_ball();
         return;
      }

      double dir = ball.m_direction;
      double dist = ball.m_distance;

      debug("GoToBall: dist="+dist+" dir="+dir+" dist_c="+ball.m_distChange+" dir_c="+ball.m_dirChange);

      /*if (dist > 5)
      {
         goToPos(estimateNextStepBallPosition());
         return;
      } */

      /*goToPos(estimateNextStepBallPosition());
      int a=1;
      if (a==1) return;*/

      int againstNear = m_memory.isNearAgainstPlayer(dir);
      dir = m_memory.findClearAngle(dir);

      double obid = ball.m_direction - dir;
      //pokus o obidenie ak je lopta daleko
      if ((Math.abs(obid)>5 && ball.m_distance>10) || play_mode!=PLAY_ON)
      {
         if (Math.abs(obid+ball.m_direction) > 2)
         {
            m_krislet.turn(obid + ball.m_direction);
            waitForNextSend();
         }
         m_krislet.dashToDist(ball.m_distance);
         return;
      }

      //uplne daleko, ignoruj male uhly...
      if (dist > 12.00)
      {
         if (Math.abs(dir)>12.00)
         {
            //if (waitForNextSend()==true) return; ------ Prvy wait uz bol v metode run() ---------
            if (Math.abs(ball.m_direction)>10.00)
            {
               m_krislet.turn(ball.m_direction);
               if (waitForNextSend()==true) return;
            }
            m_krislet.dashToDist(ball.m_distance);
         }
         else
         {
            //if (waitForNextSend()==true) return;
            m_krislet.dashToDist(ball.m_distance);
         }

      }
      else if (dist >  6.00)
      {
         if (Math.abs(dir)>8.00)
         {
            //if (waitForNextSend()==true) return;
            if (Math.abs(ball.m_direction)>8.00)
            {
               m_krislet.turn(ball.m_direction);
               if (waitForNextSend()==true) return;
            }
            m_krislet.dashToDist(ball.m_distance);
         }
         else
         {
            //if (waitForNextSend()==true) return;
            m_krislet.dashToDist(ball.m_distance);
         }

      }
      else if (dist >  3.80)
      {
         if (Math.abs(dir)>4.00)
         {
            //if (waitForNextSend()==true) return;
            if (Math.abs(ball.m_direction)>4.00)
            {
               m_krislet.turn(ball.m_direction);
               if (waitForNextSend()==true) return;
            }
            m_krislet.dashToDist(ball.m_distance);
         }
         else
         {
            //if (waitForNextSend()==true) return;
            m_krislet.dashToDist(ball.m_distance);
         }

      }
      else
      {
         //obid nepriatela
         if (againstNear!=0)
         {
            debug("Obchadzam nepriatela :-))");
            if (againstNear > 0) dir = -20.00;
            else dir = 20.00;
            //if (waitForNextSend()==true) return;
            m_krislet.turn(dir);
            if (waitForNextSend()==true) return;
            m_krislet.dash(100);

            if (ball.m_distance> 1.0)
            {
               //ball = m_memory.getObject("ball");
               if (waitForNextSend()==true) return;
               m_krislet.dash(100);
               if (waitForNextSend()==true) return;
               m_krislet.dash(100);
            }
            ball = m_memory.getObject("ball");
            if (ball!=null)
            {
               if (waitForNextSend()==true) return;
               m_krislet.turn(ball.m_direction);
            }
            else
            {
               if (waitForNextSend()==true) return;
               m_krislet.turn(-dir*2);
            }
            ball = m_memory.getObject("ball");
            if (ball!=null)
            {
               if (waitForNextSend()==true) return;
               m_krislet.dashToDist(ball.m_distance);
            }
            else
            {
               if (waitForNextSend()==true) return;
               m_krislet.dash(100);
            }


         }
         else
         {
            //ak je lopta prilis blizko ignoruj natocenie
            if (ball.m_distance<2.5 && Math.abs(ball.m_direction)<20.0)
            {
               m_krislet.dashToDist(ball.m_distance);
            }
            else if (Math.abs(dir)>3.00)
            {
               //if (waitForNextSend()==true) return;
               if (Math.abs(ball.m_direction)>3.00)
               {
                  m_krislet.turn(ball.m_direction);
                  if (waitForNextSend()==true) return;
               }
               m_krislet.dashToDist(ball.m_distance);
            }
            else
            {
               //if (waitForNextSend()==true) return;
               m_krislet.dashToDist(ball.m_distance);
            }
         }
      }
   }

   private void action_pobehuj_do_stvorca()
   {
      //if (waitForNextSend()==true) return;
      m_krislet.dash(99.99);

      if (waitForNextSend()==true) return;
      m_memory.estimateCurrentPosition();

      if (waitForNextSend()==true) return;
      m_krislet.dash(99.99);

      if (waitForNextSend()==true) return;
      m_memory.estimateCurrentPosition();

      if (waitForNextSend()==true) return;
      m_krislet.dash(99.99);

      tmpUhol+=90;
      tmpUhol = m_memory.normalize_angle(tmpUhol);

      if (waitForNextSend()==true) return;
      m_krislet.turn(tmpUhol);

      m_memory.estimateCurrentPosition();

   }

   public void goToPos(Pos tpos)
   {
      debug("Go To Pos");
      double dx= tpos.x - Krislet.pos.x;
      double dy = tpos.y - Krislet.pos.y;
      double myBodyDir=Krislet.bodyDir;
      double dist = Math.sqrt(dx * dx + dy * dy);

      //if (m_memory.isNearAgainstPlayer()) dist+=3;

      if (dx==0 && dy==0) return;

      double adir = Math.toDegrees(Math.atan2(dy, dx));

      double clear_angle = m_memory.findClearAngle();

      double turnAngle=0;

      turnAngle =  -myBodyDir + adir + clear_angle;
      boolean uzBolSend=false;
      if (dist > 12.00)
      {
         if (Math.abs(turnAngle) > 10.00)
         {
            //if (waitForNextSend()==true) return;
            m_krislet.turn(turnAngle);
            uzBolSend=true;
         }
         //if (waitForNextSend()==true) return;
         //m_krislet.dash(50);
      }
      else if (dist > 5.00)
      {
         if (Math.abs(turnAngle) > 5.00)
         {
            //if (waitForNextSend()==true) return;
            m_krislet.turn(turnAngle);
            uzBolSend=true;
         }
      }
      else if (Math.abs(turnAngle)>3.00)
      {
         //if (waitForNextSend()==true) return;
         m_krislet.turn(turnAngle);
         uzBolSend=true;
      }

      /*if (clear_angle!=0)
      {
         if (waitForNextSend()==true) return;
         m_krislet.turn(clear_angle);
         if (waitForNextSend()==true) return;
         m_krislet.dash(50);
      } */

      if (Math.abs(dist) > 1)
      {
         if (uzBolSend)
         {
            if (waitForNextSend()==true) return;
         }
         m_krislet.dashToDist(dist) ;
      }

      //debug("gotoPos: x="+tpos.x+" y="+tpos.y+" dx="+dx+" dy="+dy);
      //debug("goToPos: dist="+dist+" adir="+adir+" myBodyDir="+myBodyDir);

   }

   private boolean isObjectInHomeArea(ObjectInfo object)
   {
      Pos objPos=m_memory.estimateObjectPosition(object);
      //debug("isObjectInHomeArea: "+objPos.toString());

      if (objPos.x>home_area.getX() && objPos.x<home_area.getX()+home_area.getWidth()
            && objPos.y>home_area.getY() && objPos.y<home_area.getY()+home_area.getHeight())
               return(true);
      else return (false);
   }

   private boolean isObjectNear(ObjectInfo object, Pos pos, int tolerance)
   {
      if (object == null) return (false);
      Pos objPos=m_memory.estimateObjectPosition(object);
      if (objPos.distance(pos) < tolerance) return(true);
      else return(false);
   }

   /*private void estimateNextStepBallPosition()
   {
      if (ball.m_distance > 10) return;

      double obj_dir = ball.m_direction;

      double x1=Math.cos(Math.toRadians(obj_dir)) * ball.m_distance;
      double y1=Math.sin(Math.toRadians(obj_dir)) * ball.m_distance;

      double x2=Math.cos(Math.toRadians((double) (obj_dir + ball.m_dirChange) )) * ball.m_distance;
      double y2=Math.sin(Math.toRadians((double) (obj_dir + ball.m_dirChange) )) * ball.m_distance;

      double x3=Math.cos(Math.toRadians((double)obj_dir)) * (ball.m_distance + ball.m_distChange);
      double y3=Math.sin(Math.toRadians((double)obj_dir)) * (ball.m_distance + ball.m_distChange);

      Pos pos = new Pos();

      double multiplier = ball.m_distance /  1.43;

      //predpovedaj maximalne niekolko krokov dopredu
      if (multiplier > 1) multiplier = 1;

      double dx=x2+x3;
      double dy=y2+y3;

      dx = x1 + dx*multiplier;
      dy = y1 + dx*multiplier;

      double dist = Math.sqrt(dx * dx + dy * dy);

      //if (Math.abs(dx)<0.05 && Math.abs(dy)<0.05) return;

      double adir = Math.toDegrees(Math.atan2(dy, dx));

      debug("Updating ball pos: current dist="+ball.m_distance+" dir="+ball.m_direction);

      ball.m_distance = (float) dist;
      ball.m_direction = (float) adir;

      debug("Updating ball pos: new dist="+ball.m_distance+" dir="+ball.m_direction+" dx="+dx+" dy="+dy);
   } */

   private void estimateNextStepBallPosition()
   {
      /*debug("Ball current pos="+currentBallPos.x+","+currentBallPos.y);

      //lastBallPos.x=-999;

      if (lastBallPos.x == -999)
      {
         return;
      }
      else
      {
         //o kolko sa zmenila pozicia za posledny krok
         double dx = currentBallPos.x - lastBallPos.x;
         double dy = currentBallPos.y - lastBallPos.y;

         //debug("Updating ball pos: dx="+dx+" dy="+dy);

         if (Math.abs(dx)<1 && Math.abs(dy)<1) return;
        */
         double multiplier = (double)ball.m_distance /  1.1;

         //predpovedaj maximalne niekolko krokov dopredu
         if (multiplier > 8.0) multiplier = 8.0;
        /*
         dx = dx*multiplier;
         dy = dy*multiplier;

         //debug("Updating ball pos: dx="+dx+" dy="+dy+" multi="+multiplier);

         //pozicia voci mne
         double x1=Math.cos(Math.toRadians(Krislet.bodyDir + ball.m_direction)) * ball.m_distance;
         double y1=Math.sin(Math.toRadians(Krislet.bodyDir + ball.m_direction)) * ball.m_distance;

         dx = x1 + dx;
         dy = y1 + dy;

         double dist = Math.sqrt(dx * dx + dy * dy);

         //if (Math.abs(dx)<0.05 && Math.abs(dy)<0.05) return;

         double adir = Math.toDegrees(Math.atan2(dy, dx));
         adir = adir - Krislet.bodyDir;
        */
         debug("Updating ball pos: current dist="+ball.m_distance+" dir="+ball.m_direction);

         double krisletBodyDirChange = 0; //Krislet.bodyDir - lastKrisletBodyDir;

         //ak som sa naposledy vela otocil, nepredpokladaj
         if (Math.abs(Krislet.bodyDir - lastKrisletBodyDir)>90) multiplier = 0;

         ball.m_distance += ball.m_distChange; // * multiplier;
         ball.m_direction += (ball.m_dirChange + krisletBodyDirChange)*multiplier;

         debug("Updating ball pos: new dist="+ball.m_distance+" dir="+ball.m_direction+" body dir c="+krisletBodyDirChange+" multi="+multiplier);


      //}
   }

//===========================================================================
// Implementation of SensorInput Interface

	//---------------------------------------------------------------------------
	// This function sends see information
	public void see(VisualInfo info)
	{
		m_memory.store(info);

      //debug("Visual info="+info.toString());
	}


	//---------------------------------------------------------------------------
	// This function receives hear information from player
	public void hear(int time, int direction, String message)
	{
	}

	//---------------------------------------------------------------------------
	// This function receives hear information from referee
	public void hear(int time, String message)
	{
      debug("----> HEAR: "+message);
      try
      {
		   /*StringTokenizer tokenizer = new StringTokenizer(message,"() ", true);
		   String token;
		   // First is referee token and time token
		   tokenizer.nextToken();
		   tokenizer.nextToken();
		   tokenizer.nextToken();
		   token = tokenizer.nextToken();



		   if(token.compareTo("time_over") == 0)
			   m_timeOver = true;*/

         if (message.indexOf("goal_kick_l")!=-1) play_mode=GOAL_KICK_L;
         else if (message.indexOf("goal_kick_r")!=-1) play_mode=GOAL_KICK_R;
         else if (message.indexOf("before_kick_off")!=-1) play_mode=BEFORE_KICK_OFF;
         else if (message.indexOf("kick_off_l")!=-1) play_mode=KICK_OFF_L;
         else if (message.indexOf("kick_off_r")!=-1) play_mode=KICK_OFF_R;
         else if (message.indexOf("kick_in_l")!=-1) play_mode=KICK_IN_L;
         else if (message.indexOf("kick_in_r")!=-1) play_mode=KICK_IN_R;
         else if (message.indexOf("corner_kick_l")!=-1) play_mode=CORNER_KICK_L;
         else if (message.indexOf("corner_kick_r")!=-1) play_mode=CORNER_KICK_R;
         else if (message.indexOf("freer_kick_l")!=-1) play_mode=FREE_KICK_L;
         else if (message.indexOf("freer_kick_r")!=-1) play_mode=FREE_KICK_R;
         else if (message.indexOf("offside_l")!=-1) play_mode=OFFSIDE_L;
         else if (message.indexOf("offside_r")!=-1) play_mode=OFFSIDE_R;
         else if (message.indexOf("goal_l")!=-1) play_mode=GOAL_L;
         else if (message.indexOf("goal_r")!=-1) play_mode=GOAL_R;
         else if (message.indexOf("play_on")!=-1) play_mode=PLAY_ON;

         if (play_mode != PLAY_ON)
         {
            debug("--> GOAL - moving home...");
            //if (waitForNextSend()==true) return;
            m_krislet.move(start_pos.x, start_pos.y);

            home_pos.x = start_pos.x;
            home_pos.y = start_pos.y;

            /*if (waitForNextSend()==true) return;
            m_krislet.move(start_pos.x, start_pos.y);
            current_action = ACTION_GO_HOME;
            Krislet.pos.x = start_pos.x;
            Krislet.pos.y = start_pos.y;
            if (waitForNextSend()==true) return;
            if (waitForNextSend()==true) return;
            if (waitForNextSend()==true) return;
            if (waitForNextSend()==true) return;*/
         }
         /*if (message.indexOf("play_on")!=-1)
         {
            doPlay = true;
         }
         else
         {
            doPlay = false;
         } */
      }
      catch (Exception ex)
      {

      }
	}

   private boolean waitForNextSend()
   {
      m_memory.waitForNewInfo();
      lastKrisletBodyDir = Krislet.bodyDir;
      m_memory.estimateCurrentPosition();
      currentPos = m_memory.getCurrentPos();

      ObjectInfo o_ball = m_memory.getObject("ball");
      if (o_ball!=null)
      {
         ball = new BallInfo();
         ball.m_direction = o_ball.m_direction;
         ball.m_distance = o_ball.m_distance;

         if (Math.abs(o_ball.m_direction)>45.0)
         {
            //zachovaj stare info o zmene, nove nie je k dispozicii
         }
         else
         {
            ball.m_dirChange = o_ball.m_dirChange;
            ball.m_distChange = o_ball.m_distChange;
         }

         lastBallDir = o_ball.m_direction;

         lastBallPos.x = currentBallPos.x;
         lastBallPos.y = currentBallPos.y;

         currentBallPos = m_memory.estimateObjectPosition(ball);
         debug("Current ball pos="+currentBallPos.x+","+currentBallPos.y+" last="+lastBallPos.x+","+lastBallPos.y);

         if (ball.m_distance < 1.0 && current_action!=ACTION_KICK_BALL)
         {
            debug("Neplanovany kick :-))");
            action_kick_ball();
            return (true);
         }

         estimateNextStepBallPosition();

         //ak sa prilis meni uhol natocenia lopty, tak sa prioritne otoc namiesto
         //toho, ze niekde by si spravil dash
         if (Math.abs(ball.m_direction) > 20 && ball.m_distance>2.0)
         {
            debug("Neplanovany turn");
            m_krislet.turn(ball.m_direction);
            m_memory.waitForNewInfo();
            m_krislet.dashToDist(ball.m_distance);
            return(true);
         }

      }
      else
      {
         lastBallPos.x=-999;
      }
      return (false);

         //else neni, lebo sa moze stat ze s ball ratam (pouzijem staru...)
         /*else
         {
            //TOTO TREBA PREROBIT!!!!
            ball=null;

            lastBallPos.x = -999;
            currentBallPos.x = -999;

         } */

      //toto je povodne nezavisle casovanie...
      /*debug("WAITING --- zaciatok");
      while ((lastSendTime + sendDelay) > System.currentTimeMillis())
      {
         try
         {
            //Thread.sleep(50);
            curWaitTime = lastSendTime + sendDelay - System.currentTimeMillis();
            if (curWaitTime>10) curWaitTime=10;
            Thread.sleep(curWaitTime);
         }
         catch (Exception ex)
         {
         }
         debug("WAITING------------ T="+System.currentTimeMillis());
      }
      lastSendTime=System.currentTimeMillis();
      debug("WAITING --- koniec");*/
   }

   private void debug(String text)
   {
      if (Krislet.DEBUG) System.out.println(text);
   }

//===========================================================================
// Private members
	private SendCommand	m_krislet;			// robot which is controled by this brain
	public Memory			m_memory;				// place where all information is stored
	private char				m_side;
	volatile private boolean		m_timeOver;

   //constants
//   private double lookAngleTolerance = 7.000;
//   private double farObject = 6.000;
//   private double lookAngleToleranceFarObject = 25.000;

   //timer
   private long lastSendTime = System.currentTimeMillis();
   private long sendDelay = SoccerParams.simulator_step;
   private long curWaitTime=10;

   static final double KICK_BALL_LINE = 27;
   static final double FORWARD_LINE = 35;

   private double lastBallDir=0;

   private double lastKrisletBodyDir=0;



   //if you go home, every MaxSteps look to ball
   private int goHomeCounter=0;
   private int goHomeMaxSteps=8;

   //if play_on true
   //boolean doPlay=false;
   private int play_mode=0;
   public static final int BEFORE_KICK_OFF=0;
   public static final int KICK_OFF_L=1;
   public static final int KICK_OFF_R=2;
   public static final int KICK_IN_L=3;
   public static final int KICK_IN_R=4;
   public static final int CORNER_KICK_L=5;
   public static final int CORNER_KICK_R=6;
   public static final int GOAL_KICK_L=7;
   public static final int GOAL_KICK_R=8;
   public static final int FREE_KICK_L=9;
   public static final int FREE_KICK_R=10;
   public static final int OFFSIDE_L=11;
   public static final int OFFSIDE_R=12;
   public static final int PLAY_ON=13;
   public static final int GOAL_L=14;
   public static final int GOAL_R=15;

}