// TPerception.cpp: implementation of the TPerception class.
//
//////////////////////////////////////////////////////////////////////

#include "OS.h"
#include "stdafx.h"
#include "Player.h"
//#include "TPerceptionView.h"
#include "TParseString.h"
//#include "TViewingAgent.h"
#include "TPerception.h"
//#include "TScreenParams.h"

#include <math.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

TPerception::TPerception()
{
	CreatePerception();
}

void TPerception::CreatePerception()
{
	int	i;
	double	dFlagX[NO_OF_FLAGS] = FLAG_POSITIONS_X;
	double	dFlagY[NO_OF_FLAGS] = FLAG_POSITIONS_Y;
	
	// create the needed objects
	// players
	m_pThisPlayer = new TPlayer();	
	m_pThisPlayer->SetPlayer(0, 0.0, 0, 0.0, 0, 0, 0, TRUE, m_iSide, 0);
	for (i = 0; i < MAX_PLAYERS; i++)
		m_pVisiblePlayer[i] = new TPlayer();
	
	// goals
	m_pGoal[0] = new TSoccerGoal(-52.5, 0.0);
	m_pGoal[1] = new TSoccerGoal(52.5, 0.0);
	
	// flags
	for (i = 0; i < NO_OF_FLAGS; i++)
		m_pFlag[i] = new TFlag(dFlagX[i], dFlagY[i]);
	
	// lines
	for (i = 0; i < 5; i++)
		m_pLine[i] = new TLine();
	
	// ball
	m_pBall = new TBall();

	for(i = 0; i < MAX_FUTURE_POSITIONS; i++)
		m_pFutureBallPosition[i] = new TVector(0.0, 0.0);
	
	m_iNoOfPlayers = 0;
	m_iPlayMode = M_BEFORE_KICK_OFF;
	m_bViewLines = TRUE;
	m_iViewQuality = 1;
	m_iViewAngle = 90;

	m_dEffort = 1.0;
	m_dSpeed = 0.0;
}

TPerception::~TPerception()
{
	int	i;
	
	// delete visible objects
	delete m_pThisPlayer;
	for (i = 0; i < MAX_PLAYERS; i++)
		delete m_pVisiblePlayer[i];
	
	// delete the flags
	for (i = 0; i < NO_OF_FLAGS; i++)
		delete m_pFlag[i];
	
	// delete the lines
	for (i = 0; i < 5; i++)
		delete m_pLine[i];
	
	delete m_pGoal[0];
	delete m_pGoal[1];
	delete m_pBall;
}

/*-----------------------------------------------------------------------------
Function: void TPerception::DrawPerception(CDC* pDC, TScreenParams *tScreen)
Date    : 13.11.97
-------------------------------------------------------------------------------
Parameter: pointer to the device context to draw into
Returnval: (void)
Description: draws the last received perception
-----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
Function: void TPerception::NewMessage(LPCTSTR str)
Date    : 05.02.98
-------------------------------------------------------------------------------
Parameter: pointer to the message
Returnval: (void)
Description: calls the views function for new message,which stores it and calls 
ParseMessage
-----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
Function: BOOL TPerception::ParseMessage(LPCTSTR str)
Date    : 07.11.97
-------------------------------------------------------------------------------
Parameter: string containing the perception (max 1024 char)
Returnval: TRUE, if no parsing errors occured
Description: distinguishes between see and hear messages and calls the appropriate
specialized parsing functions
-----------------------------------------------------------------------------*/
BOOL TPerception::ParseMessage(LPCTSTR str)
{
	switch(str[1])
	{
	case 'h':
		// something to hear
		return ParseHear(str);
		
	case 's':
		// something to see or sense body
		if (str[3] == 'e')
			// a see message
			return ParseSee(str);
		else
			// a sense body information
			return ParseSenseBody(str);
		
	case 'i':
		// message at init
		return ParseInit(str);
		
	case 'r':
		// message at reconnect
		return ParseInit(str);
	}
	return TRUE;
}

/*-----------------------------------------------------------------------------
Function: BOOL TPerception::ParseSee(LPCTSTR str)
Date    : 07.11.97
-------------------------------------------------------------------------------
Parameter: reference of the string to parse
reference to the current position
Returnval: TRUE, if no parsing errors occured
Description: parses a see perception
-----------------------------------------------------------------------------*/
BOOL TPerception::ParseSee(LPCTSTR str)
{
	TParseString	msg(str);
	TParseString	strObject;
	
	// delete previously seen players
	InvalidateAll();
	
	// get the time
	msg.SetPosition(5);
	m_iSeeTime = msg.ParseInteger();
	if (msg.HasErrorOccured())
		return FALSE;
	m_pThisPlayer->SetTime(m_iSeeTime);
	
	// get the visible objects
	while (msg.ParseBracket(strObject))
	{
		if (!ParseObject(strObject))
		{
#ifdef _DEBUG
			CString error;
			error.Format(" : error parsing object info at position %d !\n", strObject.GetPosition());
			afxDump << error << strObject;
#endif
		}
	};
	
	// get the absoulte localization of the player
	Localize();
	// prepocitaj rychlost
	m_pThisPlayer->SetAbsSpeed((m_pThisPlayer->GetPosition() - m_pThisPlayer->GetOldPos()) * m_pThisPlayer->GetDecay());

	// calculate the absolute positions of the other moveable objects perceived
	GlobalMap();

	// TRUE, although may be not all objects could be parsed
	return TRUE;
}

/*-----------------------------------------------------------------------------
Function: BOOL TPerception::ParseHear(LPCTSTR str)
Date    : 10.11.97
-------------------------------------------------------------------------------
Parameter: reference to the current position in the string which is 0
Returnval: TRUE, if no parse error occured
Description: parses a hear perception
-----------------------------------------------------------------------------*/
BOOL TPerception::ParseHear(LPCTSTR str)
{
	TParseString	msg(str);
	CString			tmp;
	int				mode = 0;
	char*			strPlayMode[] = PLAYMODE_STRINGS;
	
	// get the time
	msg.SetPosition(5);
	m_iHearTime = msg.ParseInteger();
	if (msg.HasErrorOccured())
		return FALSE;
	
	// get the source
	switch (msg.GetAt(msg.GetPosition()))
	{
	case 'r':
		// it is from the referee
		if (!msg.ParseString(tmp))
			return FALSE;
		// get the message
		if (!msg.ParseString(tmp))
			return FALSE;
		
		while ((mode < NO_OF_PLAYMODES) && (tmp.Compare(strPlayMode[mode])))
			mode++;
		
		if (mode < NO_OF_PLAYMODES)
			m_iPlayMode = mode;
		break;
		
	case 's':
		// it has been sent by myself
		break;
		
	default:
		// it is a say message
		break;
	}
	
	return TRUE;
}

/*-----------------------------------------------------------------------------
Function: BOOL TPerception::ParseSenseBody(LPCTSTR str)
Date    : 27.01.98
-------------------------------------------------------------------------------
Parameter: string to parse
Returnval: FALSE, if an error occured
Description: parses a sense body information (sense_body TIME ((view_mode QUALITY
WIDTH) (stamina STAMINA EFFORT) (speed AMOUNT) (kick COUNT) (dash COUNT)
(turn COUNT) (say COUNT))
-----------------------------------------------------------------------------*/
BOOL TPerception::ParseSenseBody(LPCTSTR str)
{
	TParseString	msg(str);
	CString			temp;
	
	// read the time
	msg.SetPosition(12);
	m_iSenseTime = msg.ParseInteger();
	if (msg.HasErrorOccured())
		return FALSE;
	
	// read the view mode
	msg.ParseString(temp);
	msg.ParseString(temp);
	if (msg.HasErrorOccured())
		return FALSE;
	if (temp.Compare("low") == 0)
		m_iViewQuality = 0;
	else
		m_iViewQuality = 1;
	msg.ParseString(temp);
	if (msg.HasErrorOccured())
		return FALSE;
	if (temp.Compare("narrow") == 0)
		m_iViewAngle = 45;
	else if (temp.Compare("wide") == 0)
		m_iViewAngle = 180;
	else
		m_iViewAngle = 90;
	
	// read the stamina
	msg.ParseString(temp);
	m_iStamina = msg.ParseInteger();
	if (msg.HasErrorOccured())
		return FALSE;

	// read the effort
	m_dEffort = msg.ParseDouble();
	if (msg.HasErrorOccured())
		return FALSE;
	if (m_dEffort > 1.0)  m_dEffort = 1.0;
	if (m_dEffort < EFFORT_MIN)  m_dEffort = EFFORT_MIN;
	
	// read the speed
	msg.ParseString(temp);
	m_dSpeed = msg.ParseDouble();
	if (msg.HasErrorOccured())
		return FALSE;
	if (m_dSpeed > PLAYER_SPEED_MAX * PLAYER_DECAY)
		m_dSpeed = PLAYER_SPEED_MAX * PLAYER_DECAY;
	
	// read the counts
	msg.ParseString(temp);
	m_iNoOfKick = msg.ParseInteger();
	if (msg.HasErrorOccured())
		return FALSE;
	
	msg.ParseString(temp);
	m_iNoOfDash = msg.ParseInteger();
	if (msg.HasErrorOccured())
		return FALSE;
	
	msg.ParseString(temp);
	m_iNoOfTurn = msg.ParseInteger();
	if (msg.HasErrorOccured())
		return FALSE;
	
	msg.ParseString(temp);
	m_iNoOfSay = msg.ParseInteger();
	if (msg.HasErrorOccured())
		return FALSE;
	
	return TRUE;
}

/*-----------------------------------------------------------------------------
Function: BOOL TPerception::ParseSenseBody(LPCTSTR str)
Date    : 27.01.98
-------------------------------------------------------------------------------
Parameter: string to parse
Returnval: FALSE, if an error occured
Description: parses a sense body information (sense_body TIME ((view_mode QUALITY
				 WIDTH) (stamina STAMINA EFFORT) (speed AMOUNT) (kick COUNT) (dash COUNT)
				 (turn COUNT) (say COUNT))
-----------------------------------------------------------------------------*/
/*BOOL TPerception::ParseSenseBody(LPCTSTR str)
{
	TParseString	msg(str);
	CString			temp;
	int				time;


	// read the time
	msg.SetPosition(12);
	time = msg.ParseInteger();
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing sensetime  %s\r\n", msg);
		#endif
		return FALSE;
	}
	if (time > m_iSenseTime)
		m_bTimeProceeding = TRUE;
	else
		m_bTimeProceeding = FALSE;
	m_iSenseTime = time;

	// read the view mode
	msg.ParseString(temp);
	msg.ParseString(temp);
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing viewmode  %s\r\n", msg);
		#endif
		return FALSE;
	}
	if (temp.GetAt(0) == 'l')
		m_iViewQuality = 0;
	else
		m_iViewQuality = 1;
	msg.ParseString(temp);
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing viewquality  %s\r\n", msg);
		#endif
		return FALSE;
	}
	if (temp.GetAt(1) == 'a')
		m_iViewAngle = 45;
	else if (temp.GetAt(0) == 'w')
		m_iViewAngle = 180;
	else
		m_iViewAngle = 90;

	// read the stamina
	msg.ParseString(temp);
	SetStamina((int) msg.ParseDouble());
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing stamina  %s\r\n", msg);
		#endif
		return FALSE;
	}

	// read the effort
	m_dEffort = msg.ParseDouble();
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing effort  %s\r\n", msg);
		#endif
		return FALSE;
	}
	if (m_dEffort > 1.0)
		m_dEffort = 1.0;
	if (m_dEffort < EFFORT_MIN)
		m_dEffort = EFFORT_MIN;

	// read the speed
	msg.ParseString(temp);
	m_dSpeed = msg.ParseDouble();
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing speed  %s\r\n", msg);
		#endif
		return FALSE;
	}
	if (m_dSpeed > PLAYER_SPEED_MAX * PLAYER_DECAY)
		m_dSpeed = PLAYER_SPEED_MAX * PLAYER_DECAY;

	// read the headAngle
	msg.ParseString(temp);
	m_iHeadAngle = (int) msg.ParseDouble();
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing headangle  %s\r\n", msg);
		#endif
		return FALSE;
	}

	// read the counts
	msg.ParseString(temp);
	m_iNoOfKick = msg.ParseInteger();
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing noOfKicks  %s\r\n", msg);
		#endif
		return FALSE;
	}
	if (m_iNoOfKick < m_iExpectedNoOfKick)
	{
		// a turn has not been executed
		#ifdef _DEBUG
			TRACE0("Kick not executed\n");
		#endif
	}
	m_iExpectedNoOfKick = m_iNoOfKick;
	
	msg.ParseString(temp);
	m_iNoOfDash = msg.ParseInteger();
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing noOfDash  %s\r\n", msg);
		#endif
		return FALSE;
	}
	if (m_iNoOfDash < m_iExpectedNoOfDash)
	{
		// a dash has not been executed
		#ifdef _DEBUG
			TRACE0("Dash not executed\n");
		#endif
	}
	m_iExpectedNoOfDash = m_iNoOfDash;
	
	msg.ParseString(temp);
	m_iNoOfTurn = msg.ParseInteger();
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing noOfTurn  %s\r\n", msg);
		#endif
		return FALSE;
	}
	if (m_iNoOfTurn < m_iExpectedNoOfTurn)
	{
		// a turn has not been executed
		#ifdef _DEBUG
			TRACE0("Turn not executed\n");
		#endif
	}
	m_iExpectedNoOfTurn = m_iNoOfTurn;
	
	msg.ParseString(temp);
	m_iNoOfSay = msg.ParseInteger();
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing noOfSay  %s\r\n", msg);
		#endif
		return FALSE;
	}
	
	msg.ParseString(temp);
	m_iNoOfTurnNeck = msg.ParseInteger();
	if (msg.HasErrorOccured())
	{
		#ifdef _DEBUG
			TRACE1("Error Parsing noOfTurnNeck  %s\r\n", msg);
		#endif
		return FALSE;
	}
	if (m_iNoOfTurnNeck < m_iExpectedNoOfTurnNeck)
	{
		// a turn has not been executed
		#ifdef _DEBUG
			TRACE0("TurnNeck not executed\n");
		#endif
	}
	m_iExpectedNoOfTurnNeck = m_iNoOfTurnNeck;

//	char		buf1[200];
//	char		buf2[200];
//	double	dStamina;
//	// (sense_body 3000 (view_mode high normal) (stamina 3500 1) (speed 0) (head_angle 0) (kick 0) (dash 0) (turn 7) (say 0) (turn_neck 0))
//	int n = sscanf(str, "(sense_body %d (view_mode %s %s (stamina %f %f) (speed %f) (head_angle %d) (kick %d) (dash %d) (turn %d) (say %d) (turn_neck %d))"
//						, &m_iSenseTime, buf1, buf2, &dStamina, &m_dEffort, &m_dSpeed, &m_iHeadAngle, &m_iNoOfKick, &m_iNoOfDash, &m_iNoOfTurnNeck);
	
	// are we waiting for this result (autorun)
	if (m_bWaitForLastSenseBody)
	{
		m_bWaitForLastSenseBody = FALSE;
		m_pAgent->OnReceivedLastSenseBody();
	}

	return TRUE;
} */

/*-----------------------------------------------------------------------------
Function: BOOL TPerception::ParseInit(LPCTSTR str)
Date    : 29.01.98
-------------------------------------------------------------------------------
Parameter: string containing the init message (init side number playmode)
Returnval: FALSE, if an error occured
Description: parses the init message to get the side and players number
-----------------------------------------------------------------------------*/
BOOL TPerception::ParseInit(LPCTSTR str)
{
	TParseString	msg(str);
	CString			tmp;
	int				number;
	int				mode;
	char*			strPlayMode[] = PLAYMODE_STRINGS;
	
	// get the side
	if ((msg.GetLength() > 6) && (msg.GetAt(6) == 'r'))
		SetSide(RIGHT_TO_LEFT);
	else
		SetSide(LEFT_TO_RIGHT);
	
	// get the number of the player
	msg.SetPosition(8);
	number = msg.ParseInteger();
	if (msg.HasErrorOccured())
		return FALSE;
	SetNumber(number);
	
	// get the playmode
	msg.ParseString(tmp);
	mode = 0;
	while ((mode < NO_OF_PLAYMODES) && (tmp.Compare(strPlayMode[mode])))
		mode++;
	
	if (mode < NO_OF_PLAYMODES)
		m_iPlayMode = mode;
	
	return TRUE;
}

/*-----------------------------------------------------------------------------
Function: BOOL TPerception::ParseObject(LPCTSTR str)
Date    : 07.11.97
-------------------------------------------------------------------------------
Parameter: reference to the current position in the string, which must be the 
opening bracket of the objinfo
Returnval: TRUE, if no parse error occured
Description: parses an object and stores the information
-----------------------------------------------------------------------------*/
BOOL TPerception::ParseObject(LPCTSTR str)
{
	TParseString	msg(str);
	TParseString	name;
	CString		teamname;
	BOOL			bMyTeam;
	int			team;
	int			iNum;
	double		dist;
	int			dir;
	double		distchng;
	double		dirchng;
	int			faceDir;
	int			bodyDir;
	int			index;
	int			type;
	char			tag1;
	char			tag2;
	int			flagNum;
	int			iFlag;
	TFlag*		pFlag;
	TLine*		pLine;

	// read the objects name bracket
	if (!msg.ParseBracket(name))
		return FALSE;
	
	// read the distance, direction and distchngection (the last may be not existing)
	if (m_iViewQuality == 0)
	{
		// we have got low view quality, so no distances
		dist = NOT_EXISTING;
		dir = msg.ParseInteger();
		if (msg.HasErrorOccured())
			return FALSE;
	}
	else
	{
		// we have got high view quality
		dist = msg.ParseDouble();
		if (msg.HasErrorOccured())
			return FALSE;
		dir = msg.ParseInteger();
		if (msg.HasErrorOccured())
			return FALSE;
	}
	dirchng = NOT_EXISTING;
	distchng = msg.ParseDouble();
	if (msg.HasErrorOccured())
		distchng = NOT_EXISTING;
	else 
	{	
		dirchng = msg.ParseDouble();
		if (msg.HasErrorOccured())
			dirchng = NOT_EXISTING;
		else 
		{	
			faceDir = msg.ParseInteger();
			if (msg.HasErrorOccured())
				faceDir = NOT_EXISTING;
			else 
			{	
				bodyDir = msg.ParseInteger();
				if (msg.HasErrorOccured())
					bodyDir = NOT_EXISTING;
			}
		}
	}

	static int lastseetime = -1;
	
	// switch on the kind of object
	switch (name.GetAt(0))
	{
		// it is a player
	case 'P':break;
	case 'p':
			// read the name
			if (name.GetLength() > 7)
			{
				name.SetPosition(7);
				// read the teamname
				name.ParseString(teamname);
				// is it my team ?
				if (teamname.Compare(m_strTeam) == 0)
					team = GetSide();
				else if (GetSide() == LEFT_TO_RIGHT)
					team = RIGHT_TO_LEFT;
				else
					team = LEFT_TO_RIGHT;

				// read player number
				iNum = name.ParseInteger();
				if (name.HasErrorOccured())
					iNum = NOT_EXISTING;
			}
			else
			{
				team = NOT_EXISTING;
				iNum = NOT_EXISTING;
			} 

/*		name.SetPosition(6);
		if (name.GetLength()<7)						// no team
		{
			team = NOT_EXISTING;
			iNum = NOT_EXISTING;
		}
		else if (name.ParseString(teamname))		// read the teamname
		{
			// ist it my team ?
			if (teamname.Compare(_T(MY_TEAM)) == 0)
				team = 0;
			else
				team = 1;
			
			// read player number
			iNum = name.ParseInteger();
			if (msg.HasErrorOccured())
				iNum = NOT_EXISTING;
		}
*/
		bMyTeam = (team == m_pThisPlayer->GetTeam());
		m_pVisiblePlayer[m_iNoOfPlayers]->SetPlayer(m_iSeeTime, dist, dir, distchng
													, dirchng, faceDir, bodyDir, bMyTeam, team, iNum);
		m_iNoOfPlayers++;
		break;
	case 'G':
	case 'g':
		// it is a goal
		if ((name.GetLength() > 5) && (name.GetAt(5) == 'l'))
		{
			index = 0;
			pFlag = m_pFlag[GOAL_L];	// evidujem aj ako flag
		}
		
		else
		{
			index = 1;
			pFlag = m_pFlag[GOAL_R];	// evidujem aj ako flag
		}

		pFlag->SetVisibleObject(m_iSeeTime, dist, dir);

		// store the values
		m_pGoal[index]->SetVisibleObject(m_iSeeTime, dist, dir);
		
		break;
		
		
	case 'B':
	case 'b':
		if (lastseetime == m_iSeeTime) {
			lastseetime = m_iSeeTime;
		}
		lastseetime = m_iSeeTime;

		// it is the ball
		m_pBall->SetBall(m_iSeeTime, dist, dir, distchng, dirchng);
		break;
		
	case 'F':
		// do not treat flags by now, that are behind me
		break;
		
	case 'f':
		// it is a flag
		type = FIELD_FLAG;
		if (name.GetLength() < 6)
			return FALSE;
		tag1 = name.GetAt(5);
		if (name.GetLength() < 8)
		{
			// it must be the center flag
			if (tag1 != 'c')
				return FALSE;
			iFlag = CENTER_MARK;
		}
		else
		{
			tag2 = name.GetAt(7);
			switch (tag1)
			{	
			case 'c':
				// it is a center flag 
				if (tag2 == 't')
					iFlag = CENTER_TOP;
				else
					iFlag = CENTER_BOTTOM;
				break;
				
			case 'g':
				// it is a goal flag
				tag1 = tag2;
				if (name.GetLength() < 10)
					return FALSE;
				tag2 = name.GetAt(9);
				if (tag1 == 'l')
				{
					if (tag2 == 't')
						iFlag = GOAL_POST_L;
					else
						iFlag = GOAL_POST_L + 1;
				}
				else
				{
					if (tag2 == 't')
						iFlag = GOAL_POST_R;
					else
						iFlag = GOAL_POST_R + 1;
				}
				break;
				
			case 'p':
				// it is a flag of the penalty area
				type = PENALTY_AREA_FLAG;
				tag1 = tag2;
				if (name.GetLength() < 10)
					return FALSE;
				tag2 = name.GetAt(9);
				if (tag1 == 'l')
				{
					if (tag2 == 't')
						iFlag = FLAG_PENALTY_L;
					else if (tag2 == 'c')
						iFlag = FLAG_PENALTY_L + 1;
					else
						iFlag = FLAG_PENALTY_L + 2;
				}
				else
				{
					if (tag2 == 't')
						iFlag = FLAG_PENALTY_R;
					else if (tag2 == 'c')
						iFlag = FLAG_PENALTY_R + 1;
					else
						iFlag = FLAG_PENALTY_R + 2;
				}
				
				break;
				
			case 'l':
				// it is one of the left flags of the field
				if (tag2 == '0')
					// it is the left center flag
					iFlag = FLAG_OUTSIDE_LC;
				else
				{
					name.SetPosition(9);
					flagNum = name.ParseInteger();
					if (name.HasErrorOccured()) 
					{
						// it is an old lt or lb flag
						if (tag2 == 't') iFlag = FLAG_TL;
						else			 iFlag = FLAG_BL;
						break;
					}
					flagNum /= 10;
					flagNum--;
					if ((flagNum < 0) || flagNum > 2)
						return FALSE;
					if (tag2 == 't')
						// it is a left top outside flag
						iFlag = FLAG_OUTSIDE_LT + flagNum;
					else
						// it is a left bottom outside flag
						iFlag = FLAG_OUTSIDE_LB + flagNum;
				}
				break;
				
			case 'r':
				// it is one of the right flags of the field
				if (tag2 == '0')
					// it is the right center flag
					iFlag = FLAG_OUTSIDE_RC;
				else
				{
					name.SetPosition(9);
					flagNum = name.ParseInteger();
					if (name.HasErrorOccured()) 
					{
						// it is an old rt or rb flag
						if (tag2 == 't') iFlag = FLAG_TR;
						else			 iFlag = FLAG_BR;
						break;
					}
					flagNum /= 10;
					flagNum--;
					if ((flagNum < 0) || flagNum > 2)
						return FALSE;
					if (tag2 == 't')
						// it is a right top outside flag
						iFlag = FLAG_OUTSIDE_RT + flagNum;
					else
						// it is a right bottom outside flag
						iFlag = FLAG_OUTSIDE_RB + flagNum;
				}
				break;
				
			case 't':
				// it is one of the top flags of the field
				if (tag2 == '0')
					// it is the top center flag
					iFlag = FLAG_OUTSIDE_TC;
				else
				{
					name.SetPosition(9);
					flagNum = name.ParseInteger();
					if (name.HasErrorOccured())
						return FALSE;
					flagNum /= 10;
					flagNum--;
					if ((flagNum < 0) || flagNum > 4)
						return FALSE;
					if (tag2 == 'l')
						// it is a top left outside flag
						iFlag = FLAG_OUTSIDE_TL + flagNum;
					else
						// it is a top right outside flag
						iFlag = FLAG_OUTSIDE_TR + flagNum;
				}
				break;
				
			case 'b':
				// it is one of the bottom flags of the field
				if (tag2 == '0')
					// it is the bottom center flag
					iFlag = FLAG_OUTSIDE_BC;
				else
				{
					name.SetPosition(9);
					flagNum = name.ParseInteger();
					if (name.HasErrorOccured())
						return FALSE;
					flagNum /= 10;
					flagNum--;
					if ((flagNum < 0) || flagNum > 4)
						return FALSE;
					if (tag2 == 'l')
						// it is a bottom left outside flag
						iFlag = FLAG_OUTSIDE_BL + flagNum;
					else
						// it is a bottom right outside flag
						iFlag = FLAG_OUTSIDE_BR + flagNum;
				}
				break;
				
			default:
				iFlag = TRASH_FLAG;
				}
			}
			
			pFlag = m_pFlag[iFlag];
			pFlag->SetVisibleObject(m_iSeeTime, dist, dir);
			pFlag->SetFlag(type, tag1, tag2);
			
			break;
			
		case 'l':
			// it is a line
			if (name.GetLength() < 6)
				break;
			tag1 = name.GetAt(5);
			switch (tag1)
			{
			case 'l':
				pLine = m_pLine[0];
				break;
			case 'r':
				pLine = m_pLine[1];
				break;
			case 't':
				pLine = m_pLine[2];
				break;
			case 'b':
				pLine = m_pLine[3];
				break;
			}
			
			pLine->SetVisibleObject(m_iSeeTime, dist, dir);
			break;
	}
	
	return TRUE;
}

/*-----------------------------------------------------------------------------
Function: void TPerception::InvalidateAll()
Date    : 17.12.97
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: invalidates all visible objects. Invalidate means, they are not drawn
if they are not perceived in this cycle
-----------------------------------------------------------------------------*/
void TPerception::InvalidateAll()
{
	int	i;
	
	// ThisPlayer does not get invalidated, because it is always drawn
	// Invalidate Players
	m_iNoOfPlayers = 0;
	
	// and other objects
	for (i = 0; i < NO_OF_FLAGS; i++)
		m_pFlag[i]->Invalidate();
	
	for (i = 0; i < 5; i++)
		m_pLine[i]->Invalidate();
	
	m_pGoal[0]->Invalidate();
	m_pGoal[1]->Invalidate();
	m_pBall->Invalidate();
}

/*-----------------------------------------------------------------------------
Function: BOOL TPerception::Localize()
Date    : 16.02.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: TRUE, if the absolute position of the agent could be calculated
Description: calculates the absolute position of the agent. Two flags are 
necessary (with this algorithm) to locate the position. It is at one
of the two intersection points of the circles surounding the flags
in the distance of the player. Which intersection is decided by the 
relative angle the player sees the flags.
To calculate the intersection points, first the inbtersection point P3 of 
the line from flag1 to flag2 and the line from intersection point 1
and 2 is calculated. 

-----------------------------------------------------------------------------*/
BOOL TPerception::Localize1()
{
	int		i;
	int		iFlag1 = -1;
	int		iFlag2 = -1;
	int		iMaxAngle = -1000;
	int		iDeltaAngle;
	int		iFlag1Direction;
	double	r1, r2;				// the distance of the player to the flags 
	double	dist, dist2;		// the distance (square) of the two flags
	double	a;						// distance from one flag to intersection point P3
	double	h;						// distance from P3 to the intersection Points P1 and P2 of the two circles
	double	x, x1, x2, x3;
	double	y, y1, y2, y3;
	double	ratio;
	int		iAbsDirection;
	int		beta;

	// search for two visible flags with maximal difference of the view angle
	for (i = 0; i < NO_OF_FLAGS; i++)
	{
		if (m_pFlag[i]->IsValid() && (m_pFlag[i]->GetDistance() > 4.0))
		{
			if (iFlag1 == -1)
			{
				// found the first visible flag
				iFlag1 = i;
				iFlag1Direction = m_pFlag[iFlag1]->GetDirection();
			}
			else
			{
				iDeltaAngle = iFlag1Direction - m_pFlag[i]->GetDirection();
				if (iDeltaAngle < 0)
					iDeltaAngle = -iDeltaAngle;
				if (iDeltaAngle > iMaxAngle)
				{
					// this flag has a bigger angle to the first flag
					iFlag2 = i;
					iMaxAngle = iDeltaAngle;
				}
			}
		}
	}

	if (iFlag2 == -1)
		// there are not 2 visible flags
		return FALSE;

	// do the calculations
	r1 = m_pFlag[iFlag1]->GetDistance();
	r2 = m_pFlag[iFlag2]->GetDistance();

	x1 = m_pFlag[iFlag1]->GetAbsX();
	x2 = m_pFlag[iFlag2]->GetAbsX();

	y1 = m_pFlag[iFlag1]->GetAbsY();
	y2 = m_pFlag[iFlag2]->GetAbsY();

	mm_iFlag1 = iFlag1;
	mm_iFlag2 = iFlag2;

	// calculate the square distance of the two flags
	dist2 = (x2 - x1)*(x2 - x1) + (y2 - y1)*(y2 - y1);	
	dist = sqrt(dist2);
	if (dist > r1 + r2)
	{
		// the circles would not intersect
		dist = r1 + r2;
		dist2 = dist * dist;
	}
	else if ((r1 > r2) && (dist + r2 < r1))
	{
		// the circles would not intersect
		dist = r1 - r2;
		dist2 = dist * dist;
	}
	else if ((r2 > r1) && (dist + r1 < r2))
	{
		// the circles would not intersect
		dist = r2 - r1;
		dist2 = dist * dist;
	}

	r1 *= r1;
	r2 *= r2;
	
	// a = (r1^2 - r2^2 + d^2 ) / (2 d) 
	a = (r1 - r2 + dist2) / (2.0 * dist);
	// h^2 = r1^2 - a^2
	h = r1 - a*a;
	if (h < 0.0)
		h = 0.0;
	else
		h = sqrt(h);

	// P3 = P1 + a ( P2 - P1 ) / d
	x3 = x1 + a * (x2 - x1) / dist;
	y3 = y1 + a * (y2 - y1) / dist;

	// two circles intersect usually in 2 points. Find out which one to select
	if (m_pFlag[iFlag1]->GetDirection() <= m_pFlag[iFlag2]->GetDirection())
	{
		// result x = x3 + h ( y2 - y1 ) / d 
		x = x3 + h * (y2 - y1) / dist;
		// result y = y3 - h ( x2 - x1 ) / d 
		y = y3 - h * (x2 - x1) / dist;
	}
	else
	{
		x = x3 - h * (y2 - y1) / dist;
		y = y3 + h * (x2 - x1) / dist;
	}

	m_pThisPlayer->SetAbsX(x);
	m_pThisPlayer->SetAbsY(y);

	// calculate the absolute direction
	r1 = m_pFlag[iFlag1]->GetDistance();
	ratio = (y1 - y) / r1;
	beta = (int) (asin(ratio) * RAD2DEG);
	if (x > x1)
		iAbsDirection = 180 - beta + iFlag1Direction;
	else
		iAbsDirection = beta + iFlag1Direction;

	if (iAbsDirection > 180)  iAbsDirection -= 360;
	if (iAbsDirection < -180)  iAbsDirection += 360;

	m_pThisPlayer->SetAbsDirection(iAbsDirection);

	return TRUE;
}

BOOL TPerception::Localize2()
{
	double max_dist = 0;

	if (m_pLine[0]->IsValid() && (max_dist < m_pLine[0]->GetDistance())) 
	{	// left 
		max_dist = m_pLine[0]->GetDistance();
		if (m_pLine[0]->GetDirection()>=0)
			m_pThisPlayer->SetAbsDirection(90 + m_pLine[0]->GetDirection());
		else m_pThisPlayer->SetAbsDirection(270 + m_pLine[0]->GetDirection());
	}
	if (m_pLine[1]->IsValid() && (max_dist < m_pLine[1]->GetDistance())) 
	{	// right line
		max_dist = m_pLine[1]->GetDistance();
		if (m_pLine[1]->GetDirection()>=0)
			m_pThisPlayer->SetAbsDirection(m_pLine[1]->GetDirection()-90);
		else m_pThisPlayer->SetAbsDirection(m_pLine[1]->GetDirection()+90);
	}
	if (m_pLine[2]->IsValid() && (max_dist < m_pLine[2]->GetDistance())) 
	{	// top line
		max_dist = m_pLine[2]->GetDistance();
		if (m_pLine[2]->GetDirection()>=0)
			m_pThisPlayer->SetAbsDirection(	m_pLine[2]->GetDirection());
		else m_pThisPlayer->SetAbsDirection(m_pLine[2]->GetDirection()+180);
		
	}
	if (m_pLine[3]->IsValid() && (max_dist < m_pLine[3]->GetDistance())) 
	{	//bottom
		max_dist = m_pLine[3]->GetDistance();
		if (m_pLine[3]->GetDirection()>=0)
			m_pThisPlayer->SetAbsDirection(	m_pLine[3]->GetDirection()-180);
		else m_pThisPlayer->SetAbsDirection(m_pLine[3]->GetDirection());
	}

	double sx = 0;
	double sy = 0;
	double n  = 0;
	for (int i = 0; i < NO_OF_FLAGS; i++)
	{
		if (m_pFlag[i]->IsValid())
		{
			int dir = m_pThisPlayer->GetAbsDir() - m_pFlag[i]->GetDirection();
		//	m_pFlag[i]->GetDistance();
			sx +=  (170 - m_pFlag[i]->GetDistance()) * (m_pFlag[i]->GetAbsX() - (m_pFlag[i]->GetDistance() * cos(dir * PI/180)));
			sy +=  (170 - m_pFlag[i]->GetDistance()) * (m_pFlag[i]->GetAbsY() - (m_pFlag[i]->GetDistance() * sin(dir * PI/180)));
			n += (170- m_pFlag[i]->GetDistance());
		}
	}
	m_pThisPlayer->SetAbsX(sx/n);
	m_pThisPlayer->SetAbsY(sy/n);

	return TRUE;
}


BOOL TPerception::Localize()
{
	int		i;
	int		iFlag1 = -1;
	int		iFlag2 = -1;
	int		iFlagM = -1;
	int		iMaxAngle = -1000;
	int		iDeltaAngle;
	int		iFlag1Direction;
	double	r1, r2;				// the distance of the player to the flags 
	double	dist, dist2;		// the distance (square) of the two flags
	double	a;						// distance from one flag to intersection point P3
	double	h;						// distance from P3 to the intersection Points P1 and P2 of the two circles
	double	x, x1, x2, x3;
	double	y, y1, y2, y3;
	double	ratio;
	double	dMaxDist;
	int		iFlagDir, iAbsDirection;
	int		beta;

	// search for two visible flags with maximal difference of the view angle
	dMaxDist = 0;

	// najskor pozri OUTSIDE flags
	for (i = 7; i <= 42; i++)
	{
		if (m_pFlag[i]->IsValid() && (m_pFlag[i]->GetDistance() > 4.0)) {
			iFlag1 = i;
			iFlag1Direction = m_pFlag[i]->GetDirection();
			break;
		}
	}

	for (i = 7; i <= 42; i++)  // (i = 0; i < NO_OF_FLAGS; i++)
	{
		if (m_pFlag[i]->IsValid() && (m_pFlag[i]->GetDistance() > 4.0))
		{
			iFlagDir = m_pFlag[i]->GetDirection();
			if ((iFlag1 == -1) || (iFlag1 == i))
			{
				// found the first visible flag
				iFlag1 = i;
				iFlag1Direction = iFlagDir;
			}
			else
			{
				iDeltaAngle = iFlag1Direction - iFlagDir;
				if (iDeltaAngle < 0)
					iDeltaAngle = -iDeltaAngle;
				if (iDeltaAngle > iMaxAngle)
				{
					// this flag has a bigger angle to the first flag
					iFlag2 = i;
					iMaxAngle = iDeltaAngle;
				}
			}
			if (iFlagDir < 0) iFlagDir = -iFlagDir;
			if (dMaxDist < m_pFlag[i]->GetDistance()) {
				iFlagM = i;
				dMaxDist = m_pFlag[i]->GetDistance();
			}
		}
	}

	if (iFlag2 == -1)
		// there are not 2 visible flags
		return FALSE;

	// do the calculations
	r1 = m_pFlag[iFlag1]->GetDistance();
	r2 = m_pFlag[iFlag2]->GetDistance();

	x1 = m_pFlag[iFlag1]->GetAbsX();
	x2 = m_pFlag[iFlag2]->GetAbsX();

	y1 = m_pFlag[iFlag1]->GetAbsY();
	y2 = m_pFlag[iFlag2]->GetAbsY();

	mm_iFlag1 = iFlag1;
	mm_iFlag2 = iFlag2;

	// calculate the square distance of the two flags
	dist2 = (x2 - x1)*(x2 - x1) + (y2 - y1)*(y2 - y1);	
	dist = sqrt(dist2);
	if (dist > r1 + r2)
	{
		// the circles would not intersect
		dist = r1 + r2;
		dist2 = dist * dist;
	}
	else if ((r1 > r2) && (dist + r2 < r1))
	{
		// the circles would not intersect
		dist = r1 - r2;
		dist2 = dist * dist;
	}
	else if ((r2 > r1) && (dist + r1 < r2))
	{
		// the circles would not intersect
		dist = r2 - r1;
		dist2 = dist * dist;
	}

	r1 *= r1;
	r2 *= r2;
	
	// a = (r1^2 - r2^2 + d^2 ) / (2 d) 
	a = (r1 - r2 + dist2) / (2.0 * dist);
	// h^2 = r1^2 - a^2
	h = r1 - a*a;
	if (h < 0.0)
		h = 0.0;
	else
		h = sqrt(h);

	// P3 = P1 + a ( P2 - P1 ) / d
	x3 = x1 + a * (x2 - x1) / dist;
	y3 = y1 + a * (y2 - y1) / dist;

	// two circles intersect usually in 2 points. Find out which one to select
	if (m_pFlag[iFlag1]->GetDirection() <= m_pFlag[iFlag2]->GetDirection())
	{
		// result x = x3 + h ( y2 - y1 ) / d 
		x = x3 + h * (y2 - y1) / dist;
		// result y = y3 - h ( x2 - x1 ) / d 
		y = y3 - h * (x2 - x1) / dist;
	}
	else
	{
		x = x3 - h * (y2 - y1) / dist;
		y = y3 + h * (x2 - x1) / dist;
	}

	m_pThisPlayer->SetPosition(TVector(x, y));
	m_pThisPlayer->SetAbsSpeed(m_pThisPlayer->GetPosition() - m_pThisPlayer->GetOldPos());

	double max_dist = 0;

	if (m_pLine[0]->IsValid() && (max_dist < m_pLine[0]->GetDistance())) 
	{	// left 
		max_dist = m_pLine[0]->GetDistance();
		if (m_pLine[0]->GetDirection()>=0)
			m_pThisPlayer->SetAbsDirection(90 + m_pLine[0]->GetDirection());
		else m_pThisPlayer->SetAbsDirection(270 + m_pLine[0]->GetDirection());
	}
	if (m_pLine[1]->IsValid() && (max_dist < m_pLine[1]->GetDistance())) 
	{	// right line
		max_dist = m_pLine[1]->GetDistance();
		if (m_pLine[1]->GetDirection()>=0)
			m_pThisPlayer->SetAbsDirection(m_pLine[1]->GetDirection()-90);
		else m_pThisPlayer->SetAbsDirection(m_pLine[1]->GetDirection()+90);
	}
	if (m_pLine[2]->IsValid() && (max_dist < m_pLine[2]->GetDistance())) 
	{	// top line
		max_dist = m_pLine[2]->GetDistance();
		if (m_pLine[2]->GetDirection()>=0)
			m_pThisPlayer->SetAbsDirection(	m_pLine[2]->GetDirection());
		else m_pThisPlayer->SetAbsDirection(m_pLine[2]->GetDirection()+180);
		
	}
	if (m_pLine[3]->IsValid() && (max_dist < m_pLine[3]->GetDistance())) 
	{	//bottom
		max_dist = m_pLine[3]->GetDistance();
		if (m_pLine[3]->GetDirection()>=0)
			m_pThisPlayer->SetAbsDirection(	m_pLine[3]->GetDirection()-180);
		else m_pThisPlayer->SetAbsDirection(m_pLine[3]->GetDirection());
	}
	if (max_dist == 0) {
		// calculate the absolute direction
		r1 = m_pFlag[iFlagM]->GetDistance();
		x1 = m_pFlag[iFlagM]->GetAbsX();
		y1 = m_pFlag[iFlagM]->GetAbsY();
		ratio = (y1 - y) / r1;
		beta = (int) (asin(ratio) * RAD2DEG);
		if (x > x1)
			iAbsDirection = 180 - beta + m_pFlag[iFlagM]->GetDirection();
		else
			iAbsDirection = beta + m_pFlag[iFlagM]->GetDirection();
		
		if (iAbsDirection > 180)  iAbsDirection -= 360;
		if (iAbsDirection < -180)  iAbsDirection += 360;
		
		m_pThisPlayer->SetAbsDirection(iAbsDirection);
	}
	
	return TRUE;
}

/*-----------------------------------------------------------------------------
Function: void TPerception::GlobalMap()
Date    : 18.05.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: calculates the absolute position of all seen moveable objects
-----------------------------------------------------------------------------*/
void TPerception::GlobalMap()
{
	int	i;
	int	dir = m_pThisPlayer->GetAbsDir();
	
	// set balls global coordinates
	if (m_pBall->IsValid())
		m_pBall->SetAbsCoordinates(m_pThisPlayer, dir);

	// calculate the new future positions of the ball
	for (i = 0; i < MAX_FUTURE_POSITIONS; i++)
		m_pBall->GetFuturePosition(i, *m_pFutureBallPosition[i]);
	
	for (i = 0; i < m_iNoOfPlayers; i++)
		if (m_pVisiblePlayer[i]->IsValid())
			m_pVisiblePlayer[i]->SetAbsCoordinates(m_pThisPlayer, dir);

//	for (i = 0; i < 1; i++)
//		if (m_pGoal[i]->IsValid())
//			m_pGoal[i]->SetAbsCoordinates(m_pThisPlayer, dir);
}

/*-----------------------------------------------------------------------------
Function: Functions returning properties of the world
Date    : 24.02.98
-------------------------------------------------------------------------------
Parameter: none
Returnval: property
Description: 
-----------------------------------------------------------------------------*/
int TPerception::GetGoalDirection()
{
	int	index = 0;
	
	// which one is the correct goal
	if (GetSide() == LEFT_TO_RIGHT)
		index = 1;
	
	// if the player can see the goal, then return visible direction
	if (m_pGoal[index]->IsValid())
		return m_pGoal[index]->GetDirection(); 
	else
		// the goal is not visible, so use the absolute position of the agent
		return m_pThisPlayer->GetAngleTo(m_pGoal[index]);
}

double TPerception::GetGoalDistance()
{
	int	index = 0;
	
	// which one is the correct goal
	if (GetSide() == LEFT_TO_RIGHT)
		index = 1;
	
	return m_pThisPlayer->GetDistanceTo(m_pGoal[index]);
}

TSoccerGoal* TPerception::GetGoal()
{
	int	index = 0;
	
	// which one is the correct goal
	if (GetSide() == LEFT_TO_RIGHT)
		index = 1;
	
	return m_pGoal[index];
}

TSoccerGoal* TPerception::GetOurGoal()
{
	int	index = 1;
	
	// which one is the correct goal
	if (GetSide() == LEFT_TO_RIGHT)
		index = 0;
	
	return m_pGoal[index];
}
