// TPerception.h: interface for the TPerception class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_TPERCEPTION_H__DA8AA141_577A_11D1_B242_0000C09B5CBE__INCLUDED_)
#define AFX_TPERCEPTION_H__DA8AA141_577A_11D1_B242_0000C09B5CBE__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000


#include "TVisibleObject.h"

//class TScreenParams;
//class TPerceptionView;
//class TViewingAgent;

/*-----------------------------------------------------------------------------
Class   : TPerception
Date    : 07.11.97
-------------------------------------------------------------------------------
Base Class: CObject

Description: contains the methods for parsing the perceptions
-----------------------------------------------------------------------------*/
#define MAX_PLAYERS 22

// no of stored future positions of the ball
#define MAX_FUTURE_POSITIONS	20

class TPerception : public CObject  
{
	public:
		friend class CPlayerView;
		friend class CPlayerDoc;
TPerception();
		virtual ~TPerception();
		
		void	CreatePerception();
		
		//void	NewMessage(LPCTSTR msg);
		BOOL	ParseMessage(LPCTSTR msg);

		// Attribute functions
		void	SetPlayMode(int mode) { m_iPlayMode = mode; };
		int		GetPlayMode(void) const { return m_iPlayMode; };
		void	SetViewLines(BOOL view) { m_bViewLines = view; };
		BOOL	GetViewLines(void) const { return m_bViewLines; };
		
		// proposition functions 
		int		GetBallDirection() const { return m_pBall->GetDirection(); };
		double	GetBallDistance() const { return m_pBall->GetDistance(); };
		int		GetGoalDirection();
		double	GetGoalDistance();
		TSoccerGoal* GetGoal();
		TSoccerGoal* GetOurGoal();
		BOOL		GetSeeBall() const { if (m_iSeeTime == m_pBall->GetTime()) return TRUE; else return FALSE; };

		void	SetViewParams(int angle, int quality) { m_iViewAngle = angle; m_iViewQuality = quality; };
		int		GetViewQuality(void) const { return m_iViewQuality; };
		int		GetViewAngle(void) const { return m_iViewAngle; };
		TPlayer*	GetThisPlayer(void) const { return m_pThisPlayer; };
		int		GetSide() { return m_iSide; };
		void	SetSide(int side) { m_iSide = side; m_pThisPlayer->SetTeam(TRUE, side); };
		int		GetNumber() { return m_iNumber; };
		void	SetNumber(int number) { m_iNumber = number; };

		double	GetEffort() const { return m_dEffort; };

		int GetNumOfVisiblePlayers() { return m_iNoOfPlayers; };
        TBall*	GetBall(void) const { return m_pBall; };
		TFlag** GetFlags()  const { return (TFlag**)m_pFlag; };
		TPlayer** GetVisiblePlayers() const { return (TPlayer**)m_pVisiblePlayer; };
		TVector*	GetFutureBallPosition(int i) { ASSERT((i >= 0) && (i < MAX_FUTURE_POSITIONS));
						return m_pFutureBallPosition[i]; };

		int GetNumOfRegPlayers() { return m_iNoOfPlayers; };
		TPlayer** GetRegPlayers() const { return (TPlayer**)m_pVisiblePlayer; };

		int			mm_iFlag1, mm_iFlag2;
		float		mm_x1,mm_x2,mm_y1,mm_y2;

	private:

		BOOL		ParseSee(LPCTSTR str);
		BOOL		ParseHear(LPCTSTR str);
		BOOL		ParseSenseBody(LPCTSTR str);
		BOOL		ParseInit(LPCTSTR str);
		BOOL		ParseObject(LPCTSTR str);
		void		InvalidateAll(void);
		void		Insert(CObList &list, TVisibleObject* obj);
		BOOL		Localize(void);
		BOOL		Localize1(void);
		BOOL		Localize2(void);
		void        GlobalMap();
		
		int			m_iSeeTime;				// time received from the last perception
		int			m_iHearTime;
		int			m_iSenseTime;			// time received from the last sense_body
		
		int			m_iPlayMode;			// mode the game is in (e.g. before_kickoff)
		int			m_iNoOfPlayers;			// number of currently visible players
		TPlayer*	m_pVisiblePlayer[MAX_PLAYERS];	// max 22 players
		TPlayer*	m_pThisPlayer;
		
		TSoccerGoal* m_pGoal[2];		// 0 is the own goal
		TBall*		m_pBall;			// the ball
		TFlag*		m_pFlag[NO_OF_FLAGS];		// flags of the corners and midline
		TLine*		m_pLine[5];

		TVector*	m_pFutureBallPosition[MAX_FUTURE_POSITIONS];
		
		BOOL		m_bViewLines;
		// player info
		int			m_iSide;
		int			m_iNumber;
		
		// sense body information
		int		m_iViewQuality;
		int		m_iViewAngle;
		int		m_iStamina;
		double	m_dEffort;
		double	m_dSpeed;
		int		m_iNoOfKick;
		int		m_iNoOfDash;
		int		m_iNoOfTurn;
		int		m_iNoOfSay;

	public:
		CString	   m_strTeam;	// name of the team the player is in
};


#endif // !defined(AFX_TPERCEPTION_H__DA8AA141_577A_11D1_B242_0000C09B5CBE__INCLUDED_)
